// this DIVERSE DSO reads in a path file as created by the DTK playbackWrite
// DSO and creates a series of screen images that comprise a movie of the
// path.  It writes the frames in pseudo-real-time, i.e. it might take
// longer to create the movie than it takes to play the movie, but the movie
// will take as long to play as the path assuming the movie player can keep
// up with the generated frame rate

// The window that's dumped to the files is pipe(0)->pwin(0)

// The screen area to be dumped is at window's original size and location.
// Move or resize it and you'll get whatever is being displayed where the
// window originally was

// The image is made by dumping the area covered by the
// dpfDisplay::masterScreen()->pfchan(0) to sequentially numbered files.

// the input file name is specified by the environment variable
// DTK_NAV_PLAYBACK_FILE, if set, otherwise the file used is
// "/tmp/playback-$LOGNAME"

// the base of the output file name is specified by the environment variable
// DPF_PLAYBACK_MOVIE_FILE, if set, otherwise the base of the file used is
// "/tmp/playbackMovie-$LOGNAME".  In either case the base has a six digit
// number appended to it, and the extension "rgb".

// Author: John Kelso, kelso@nist.gov, August 2003

//      Scientific Applications & Visualization Group
//           SAVG web page: http://math.nist.gov/mcsd/savg/
//      Mathematical & Computational Science Division
//           MCSD web page: http://math.nist.gov/mcsd/
//      National Institute of Standards and Technology
//           NIST web page: http://www.nist.gov/

// "This software was developed at the National Institute of Standards
// and Technology by employees of the Federal Government in the course
// of their official duties. Pursuant to title 17 Section 105 of the
// United States Code this software is not subject to copyright
// protection and is in the public domain.  xxxxxxx is an experimental 
// system.  NIST assumes no responsibility whatsoever for its use by 
// other parties, and makes no guarantees, expressed or implied, about 
// its quality, reliability, or any other characteristic.
// 
// We would appreciate acknowledgement if the software is used."

// amount of pseudo-realtime to advance each frame
#define FRAMETIME (1.f/30.f)

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dpf.h>
#include <dtk/dtkDSO_loader.h>

class movieFromPathfile : public dtkNav {

public:

  movieFromPathfile (dpf *app) ;
  ~movieFromPathfile (void) ;
  int preConfig (void) ;
  int postConfig (void) ;
  int postFrame (void) ;

private:
  
  FILE* pathfile ;
  char* moviefilebase ;
  
  dpf *app ;
  pfPipeWindow *pfpwin ;
  int xo, yo, xs, ys ;

  dtkDequeuer *dequeuer;
  dtkInXKeyboard *keyboard;

  dtkCoord  prevPathfileLoc ;
  dtkCoord  currPathfileLoc ;

  long double currPathfileTime ;
  long double prevPathfileTime ;
  long double frameTime ;

  KeyCode xkc_l;

};

////////////////////////////////////////////////////////////////////////
movieFromPathfile::movieFromPathfile (dpf *app_in) :
  dtkNav(app_in, "movieFromPathfile") {
  app = app_in ;
  setDescription("creates a flip-book movie from a DTK path file") ;
  
  pfpwin = NULL ;
  keyboard = NULL;
  dequeuer = NULL;

  // this code modifed from DTK's code- assumes DTK leaves off the ".log"
  char *pathfilename ;
  // do this so freeing pathname won't cause a core dump
  if (getenv("DTK_NAV_PLAYBACK_FILE")) {  
    pathfilename = strdup(getenv("DTK_NAV_PLAYBACK_FILE")) ;
  } else {
    char *format = "/tmp/playback-%s" ;
    char *logname = getenv("LOGNAME") ;
    if(!logname) logname = "nonknown_user" ;
    pathfilename = (char *) malloc(strlen(format) + strlen(logname)) ;
    sprintf(pathfilename, format, logname) ;
  }

  if( (pathfile = fopen(pathfilename,"r")) == NULL) {
    dtkMsg.add(DTKMSG_ERROR,
	       "movieFromPathfile::movieFromPathfile() failed "
	       "to open pathfilename \"%s\".\n",
	       pathfilename) ;
    free(pathfilename) ;
    return ;
  }
  
  dtkMsg.add(DTKMSG_NOTICE, "movieFromPathfile: reading path data from %s\n",
	     pathfilename) ;
  free(pathfilename) ;
  
  if(getenv("DPF_PLAYBACK_MOVIE_FILE")) {
    moviefilebase = strdup(getenv("DPF_PLAYBACK_MOVIE_FILE")) ;
  } else {
    char *format = "/tmp/playbackMovie-%s" ;
    char *logname = getenv("LOGNAME") ;
    if(!logname) logname = "nonknown_user" ;
    moviefilebase = (char *) malloc(strlen(format) + strlen(logname)) ;
    sprintf(moviefilebase, format, logname) ;
  }

}

////////////////////////////////////////////////////////////////////////
movieFromPathfile::~movieFromPathfile (void) {
  
  if (pathfile) {
    fclose(pathfile) ;
    pathfile = NULL ;
  }

  if (moviefilebase) {
    free(moviefilebase) ;
    moviefilebase = NULL ;
  }

  if (dequeuer) {
    delete dequeuer ;
    dequeuer = NULL ;
  }
  
}

////////////////////////////////////////////////////////////////////////
// read in initial data
int movieFromPathfile::preConfig(void) {
  dtkMsg.add(DTKMSG_NOTICE,"movieFromPathfile: setting pfMultiprocess(0)\n") ;
  pfMultiprocess(0) ;
  // dtkNav doesn't have a preConfig callback
  return dtkAugment::REMOVE_CALLBACK ;
  //return dtkNav::preConfig() ;
}
  
////////////////////////////////////////////////////////////////////////
// read in initial data
int movieFromPathfile::postConfig(void) {

  dequeuer = new dtkDequeuer(app->record());
  if(!dequeuer || dequeuer->isInvalid()) return ERROR_;
  
  keyboard = static_cast<dtkInXKeyboard *>(app->get("xkeyboard", 
						    DTKINXKEYBOARD_TYPE));
  if(!keyboard) return ERROR_;
  keyboard->queue();

  if(!(xkc_l = keyboard->getKeyCode(XK_l)))
    return ERROR_;
  
  // get the window to dump
  pfpwin = app->display()->first()->first()->pfpwin() ;

  if (!pfpwin) {
    dtkMsg.add(DTKMSG_ERROR,1,
	       "movieFromPathfile::postConfig() failed "
	       "to find pipe(0)->pwin(0)->pfpwin().\n") ;
    return dtkAugment::ERROR_ ;
  }

  // initialize location and time;
  fscanf (pathfile, "%Lg %f %f %f %f %f %f", &currPathfileTime,
	  &(currPathfileLoc.x), &(currPathfileLoc.y),
	  &(currPathfileLoc.z), &(currPathfileLoc.h),
	  &(currPathfileLoc.p), &(currPathfileLoc.r)) ;
  
#if 0
  printf("initial time and location:\n") ;
  printf("  %Lf, (%f %f %f) (%f %f %f)\n",
	 currPathfileTime,
	 currPathfileLoc.x, currPathfileLoc.y, currPathfileLoc.z, 
	 currPathfileLoc.h, currPathfileLoc.p, currPathfileLoc.r) ;
#endif

  prevPathfileLoc = currPathfileLoc ;
  frameTime = prevPathfileTime = currPathfileTime ;
  
  dtkMsg.add(DTKMSG_NOTICE,"movieFromPathfile: move the window to the desired location \n") ;
  dtkMsg.add(DTKMSG_NOTICE,"    and size, then press the \"L\" key\n") ;
  dtkMsg.add(DTKMSG_NOTICE,"    to create the frame files.\n\n") ;
  dtkMsg.add(DTKMSG_NOTICE,"    ONCE RECORDING, DON'T MOVE OR RESIZE THE WINDOW!!\n") ;
    
  return dtkNav::postConfig() ;
}


////////////////////////////////////////////////////////////////////////
// dump the current "frame" and set the navigation DCS for next "frame"
int movieFromPathfile::postFrame (void) {

  // frame counter
  static int count = 0 ;

  // flags whether we're ready to go or not
  static int ready = 0 ;

  // for getting the "L" key
  dtkRecord_event* e;
  KeyCode key;

  if (!ready) {
    e=dequeuer->getNextEvent(keyboard) ;
    if (e) {
      keyboard->read(&key,e) ;
      if (key != xkc_l)
	return dtkNav::postFrame() ;
      else {
	pfpwin->getCurScreenOriginSize(&xo, &yo, &xs, &ys) ; 
	ready = 1 ;
      }
    } else
      return dtkNav::postFrame() ;
  }

  pfpwin->getCurScreenOriginSize(&xo, &yo, &xs, &ys) ;

  // file name of snap-shot of this frame
  static char *frameFileName = NULL ;

  // advance pseudo-realtime counter
  frameTime += FRAMETIME ;

  // read file until frame time is past path time
  while(frameTime >= currPathfileTime) {

    prevPathfileLoc   = currPathfileLoc ;
    prevPathfileTime = currPathfileTime ;
    
    // reads from file, if there is nothing there prints and unloads
    int n = fscanf (pathfile, "%Lg%f%f%f%f%f%f", &currPathfileTime,
		    &(currPathfileLoc.x), &(currPathfileLoc.y),
		    &(currPathfileLoc.z), &(currPathfileLoc.h),
		    &(currPathfileLoc.p), &(currPathfileLoc.r)) ;
#if 0
    printf("%d items read at frame time %Lf\n", n, frameTime) ;
    printf("  %Lf, (%f %f %f) (%f %f %f)\n",
	   currPathfileTime,
	   currPathfileLoc.x, currPathfileLoc.y, currPathfileLoc.z, 
	   currPathfileLoc.h, currPathfileLoc.p, currPathfileLoc.r) ;
#endif    
    if (7!=n) {
      if (frameFileName) {
	dtkMsg.add(DTKMSG_NOTICE,"movieFromPathfile: "
		   "last move frame filename = \"%s\"\n",
		   frameFileName) ;
	free(frameFileName) ;
      } else {
	dtkMsg.add(DTKMSG_NOTICE,"movieFromPathfile: "
		   "No frame files written\n") ;
      }
      // this'll cause the program to exit gracefully if the app is checking
      // for it
      app->state &= ~(DTK_ISRUNNING);
      return dtkAugment::REMOVE_OBJECT;
    }
  }
  
  // process this frame of data

  // roll up a file name for this frame
  if (frameFileName) {
    free(frameFileName) ;
    frameFileName = NULL ;
  }
  frameFileName = (char *) malloc(strlen(moviefilebase) + 12) ;
  sprintf(frameFileName,"%s-%6.6d.rgb",moviefilebase,count) ;
  
  if (count == 0)
    dtkMsg.add(DTKMSG_NOTICE,"movieFromPathfile: "
	       "first move frame filename = \"%s\"\n",
	       frameFileName) ;
  

  dtkMsg.add(DTKMSG_INFO,"movieFromPathfile: time = %Lf, filename = %s\n",
	     frameTime, frameFileName) ;
  
  dtkMsg.add(DTKMSG_NOTICE,"frame %8.8d               \r", count) ;
  fflush(stdout) ;

  count++ ;
  
  pfuSaveImage(frameFileName, xo, yo, xs, ys, 0) ;
  
  // linear interpolates data
  location.x = prevPathfileLoc.x+(frameTime-prevPathfileTime)*
    (currPathfileLoc.x-prevPathfileLoc.x)/(currPathfileTime-prevPathfileTime) ;
  location.y = prevPathfileLoc.y+(frameTime-prevPathfileTime)*
    (currPathfileLoc.y-prevPathfileLoc.y)/(currPathfileTime-prevPathfileTime) ;
  location.z = prevPathfileLoc.z+(frameTime-prevPathfileTime)*
    (currPathfileLoc.z-prevPathfileLoc.z)/(currPathfileTime-prevPathfileTime) ;
  location.h = prevPathfileLoc.h+(frameTime-prevPathfileTime)*
    (currPathfileLoc.h-prevPathfileLoc.h)/(currPathfileTime-prevPathfileTime) ;
  location.p = prevPathfileLoc.p+(frameTime-prevPathfileTime)*
    (currPathfileLoc.p-prevPathfileLoc.p)/(currPathfileTime-prevPathfileTime) ;
  location.r = prevPathfileLoc.r+(frameTime-prevPathfileTime)*
    (currPathfileLoc.r-prevPathfileLoc.r)/(currPathfileTime-prevPathfileTime) ;
  
  return dtkNav::postFrame() ;
}

/************ DTK C++ dtkAugment loader/unloader functions ***************
 *
 * All DTK dtkAugment DSO files are required to declare these two
 * functions.  These function are called by the loading program to get
 * your C++ objects loaded.
 *
 *************************************************************************/

static dtkAugment *dtkDSO_loader(dtkManager *manager, void *p)
{
  dpf *app = static_cast<dpf *>(manager) ;
  return new movieFromPathfile(app) ;
}

static int dtkDSO_unloader(dtkAugment *augment)
{
  delete augment ;
  return DTKDSO_UNLOAD_CONTINUE ;
}


